# frozen_string_literal: true
require "pry"
require "yaml"

module Wilco
  class Chatbot
    def self.persisted(store, id)
      bot = new.load_from(store, id)
      yield(bot)
      bot.save_to(store, id)
    end

    def initialize(options = {})
      @commands = options.fetch(:commands)
      @messages = options.fetch(:messages)
      @parser = options.fetch(:parser, Parser.new)
    end

    def greet
      "#{t(:greeting)} #{t(:command_prompt)} #{t(:suggestions)}"
    end

    def goodbye
      t(:valediction)
    end

    def respond_to_message(message)
      if current_command
        identify_param(message)
      else
        identify_command(message)
      end
    end

    def save_to(store, id)
      serialized_command = YAML.dump(current_command)
      store.set(serialization_key(id), serialized_command)
    end

    def load_from(store, id)
      serialized_command = store.get(serialization_key(id))
      @current_command = YAML.load(serialized_command) if serialized_command
      self
    end

    private

    attr_reader :commands, :messages, :current_command, :parser

    def t(key)
      messages[key]
    end

    def serialization_key(id)
      "intent_chatbot_#{id}"
    end

    def identify_command(message)
      if (@current_command = command_for(message))
        respond_to_current_command_state
      else
        did_not_understand_command
      end
    end

    def did_not_understand_command
      "#{t(:did_not_understand)} #{t(:command_prompt)} #{t(:suggestions)}"
    end

    def did_not_understand_param
      "#{t(:did_not_understand)} " \
      "#{current_command.request_first_missing_param}"
    end

    def identify_param(message)
      if current_command.extract_param(message)
        respond_to_current_command_state
      elsif (param_name = current_command.find_param_to_switch_to(message))
        current_command.request_param(param_name)
      else
        did_not_understand_param
      end
    end

    def respond_to_current_command_state
      if current_command.missing_params?
        current_command.request_first_missing_param
      else
        response = current_command.execute
        @current_command = nil
        "#{response} #{t(:next_command_prompt)}"
      end
    end

    def command_for(message)
      command = command_instances.max_by { |c| c.likelihood(message) }
      command.parse_message(message) if command.likelihood(message).positive?
    end

    def command_instances
      @command_instances ||= commands.map { |command_class|
        command_class.new(parser)
      }
    end
  end
end
